/* -*- Mode: c++; -*- */
/*  --------------------------------------------------------------------
 *  Filename:
 *    src/bincimap-up.cc
 *  
 *  Description:
 *    Implementation of the preauthenticated bincimap stub
 *
 *  Authors:
 *    Andreas Aardal Hanssen <bincimap@andreas.hanssen.name>
 *
 *  Bugs:
 *
 *  ChangeLog:
 *
 *  --------------------------------------------------------------------
 *  Copyright 2002, 2003 Andreas Aardal Hanssen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *  --------------------------------------------------------------------
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string>

#include <signal.h>

#include "broker.h"
#include "depot.h"
#include "exceptions.h"
#include "io/io.h"
#ifdef WITH_SSL
#include "io/io-ssl.h"
#endif
#include "mailbox/maildir/maildir.h"
#include "operators.h"
#include "session.h"

using namespace ::std;
using namespace Binc;

namespace Binc {
  bool showGreeting(void);
  void initConfig(void);
}

namespace {
  void alarmhandler(int signum)
  {
    IO &com = IOFactory::getInstance().get(1);
    IO &logger = IOFactory::getInstance().get(2);

    logger << "Connection closed (timed out)" << endl;
    com << "* BYE Binc IMAP shutting down - timeout exceeded" << endl;
    com.flushContent();
    logger.flushContent();
    exit(0);
  }

  //------------------------------------------------------------------------
  void usage(char *name)
  {
    IO &logger = IOFactory::getInstance().get(2);

    logger << "Please refer to the man pages for bincimap-up and bincimapd" << endl;
    logger << "for more information about how to invoke Binc IMAP." << endl;
    logger.flushContent();
  }
}

//------------------------------------------------------------------------
int main(int argc, char **argv) 
{
  IOFactory &iofactory = IOFactory::getInstance();

#ifdef WITH_SSL
  SSLEnabledIO *sslio = new SSLEnabledIO(stdout);
  iofactory.assign(1, sslio);
#else
  iofactory.assign(1, new IO(stdout));
#endif

  iofactory.assign(2, new IO(stderr));

  IO &com = iofactory.get(1);
  IO &logger = iofactory.get(2);

  Session &session = Session::getInstance();

  logger.noFlushOnEndl();

  // Read command line arguments
  try {
    session.parseCommandLine(argc, argv);
  } catch (SyntaxException &e) {
    logger << "Error reading command line arguments, "
	   << e.getReason() << endl;
    logger.flushContent();
    return 111;
  }

  // Show help if asked for it
  if (session.command.help) {
    usage(argv[0]);
    logger.flushContent();
    return 111;
  }

  // Show help if asked for it
  if (session.command.version) {
    logger << "Binc IMAP v" << VERSION << endl;
    logger.flushContent();
    return 111;
  }

  // Grab config file name
  string configfile = session.command.configfile;

  // Read configuration
  if (configfile != "")
    try {
      session.globalconfig.load(configfile);
    } catch (StorageException &e) {
      logger << "Error reading global config, " << e.getReason() << endl;
      logger.flushContent();
      return 111;
    }

  // Read configuration settings
  session.initConfig();
  DepotFactory &depotfactory = DepotFactory::getInstance();
  depotfactory.assign(new IMAPdirDepot());
  depotfactory.assign(new MaildirPPDepot());

  session.globalconfig.setSection("Mailbox");
  string depottype = session.globalconfig["depot"];
  if (depottype == "") depottype = "Maildir++";

  Depot *dep = depotfactory.get(depottype);
  if (dep == NULL) {
    logger << "Found no Depot for \"" << depottype << "\""
	   << endl;
    logger.flushContent();
    return 111;
  }

  Depot &depot = *dep;

  BrokerFactory brokerfactory = BrokerFactory::getInstance();

  brokerfactory.assign(new AuthenticateOperator());
  brokerfactory.assign(new CapabilityOperator());
  brokerfactory.assign(new LoginOperator());
  brokerfactory.assign(new LogoutOperator());
  brokerfactory.assign(new NoopOperator());
#ifdef WITH_SSL
  brokerfactory.assign(new StarttlsOperator());
#endif

  // log settings
  session.globalconfig.setSection("Log");
  string ipenv = session.globalconfig["ip environment variable"];
  // Initialize logger
  string ip = getenv(ipenv.c_str()) ? getenv(ipenv.c_str()) :
    getenv("TCPREMOTEIP") ? getenv("TCPREMOTEIP") :
    getenv("REMOTE_HOST") ? getenv("REMOTE_HOST") :
    getenv("REMOTEIP") ? getenv("REMOTEIP") :
    getenv("SSLREMOTEIP") ? getenv("SSLREMOTEIP") : "?";
  session.setIP(ip);

  if (session.globalconfig["type"] == "multilog" 
      || session.globalconfig["type"] == "stderr") {
    logger.setLogPrefix("unknown@" + ip + ":");
    logger.enableLogPrefix();
  } else if (session.globalconfig["type"] == "" 
	     || session.globalconfig["type"] == "syslog") {
    session.globalconfig.setSection("Log");
    const string f = session.globalconfig["syslog facility"];
    const string fn = session.globalconfig["syslog facility number"];

    int facility;

    if (fn != "") facility = atoi(fn);
    else {
      if (f == "LOG_USER") facility = LOG_USER;
      else if (f == "LOG_LOCAL0") facility = LOG_LOCAL0;
      else if (f == "LOG_LOCAL1") facility = LOG_LOCAL1;
      else if (f == "LOG_LOCAL2") facility = LOG_LOCAL2;
      else if (f == "LOG_LOCAL3") facility = LOG_LOCAL3;
      else if (f == "LOG_LOCAL4") facility = LOG_LOCAL4;
      else if (f == "LOG_LOCAL5") facility = LOG_LOCAL5;
      else if (f == "LOG_LOCAL6") facility = LOG_LOCAL6;
      else if (f == "LOG_LOCAL7") facility = LOG_LOCAL7;
      else facility = LOG_DAEMON;
    }

    session.globalconfig.insert("syslog facility number", toString(facility));

    logger.setModeSyslog("bincimap-up", facility);
  }

  // Now that we know the log type, we can flush.
  logger.flushContent();
  logger.flushOnEndl();

#ifdef WITH_SSL
  // Set SSL mode if --ssl is passed
  if (session.command.ssl)
    try {
      sslio->setModeSSL();
      session.add("sslmode", "true");
    } catch (InputException &e) {
      logger << "SSL negotiation failed: " << e.getReason() << endl;
      return 111;
    }
#endif

  session.setState(Session::NONAUTHENTICATED);

  // Read timeout settings from global config
  session.globalconfig.setSection("Session");
  int idletimeout = atoi(session.globalconfig["idle timeout"]);
  int authtimeout = atoi(session.globalconfig["auth timeout"]);
  int transfertimeout = atoi(session.globalconfig["transfer timeout"]);

  // Settings are not allowed to break the IMAP protocol
  if (idletimeout < (30 * 60)) idletimeout = 30 * 60;

  // No auth timeout is not allowed.
  if (authtimeout < 30) authtimeout = 30;

  // Set transfer timeout
  com.setTransferTimeout(transfertimeout);
  logger.setTransferTimeout(transfertimeout);

  // Read transfer buffer size
  int buffersize = atoi(session.globalconfig["transfer buffer size"]);
  com.setBufferSize(buffersize >= 0 ? buffersize : 0);

  // umask settings
  session.globalconfig.setSection("Mailbox");
  string umsk = session.globalconfig["umask"];
  if (umsk == "" || !isdigit(umsk[0])) umsk = "0777";
  unsigned int mode;
  sscanf(session.globalconfig["umask"].c_str(), "%o", &mode);
  umask((mode_t) mode);

  // Initialize timeout handler
  signal(SIGALRM, alarmhandler);

  // Show standard greeting
  if (!showGreeting())
    return 111;

  logger << "Client connected to Binc IMAP from " << session.getIP() << endl;

  // Read commands and run functions
  while (session.getState() != Session::LOGOUT) {
    Command command;
    bool recovery = false;
    try {
      Broker &broker = brokerfactory.getBroker(session.getState());
      
      com.noFlushOnEndl();

      alarm(session.getState() == Session::NONAUTHENTICATED 
	    ? authtimeout
	    : idletimeout);

      com.enableInputLimit();
      broker.parse(command);
      alarm(0);

      broker.process(depot, command);

      com << command.getTag() << " OK " << command.getName() 
	  << " completed" << endl;

    } catch (CriticalException &e) {
      logger << "Critical exception: " << e.getReason() << endl;
      session.setState(Session::LOGOUT);

    } catch (OperatorException &e) {
      if (e.getCode() != Operator::NOTHING) {
	com << command.getTag();

	if (e.getCode() == Operator::OK) com << " OK ";
	else if (e.getCode() == Operator::BAD) { com << " BAD "; }
	else if (e.getCode() == Operator::NO) com << " NO ";
	com << e.getReason() << endl;
      }

    } catch (StorageException &e) {
      logger << "Storage warning, shutting down: " << e.getReason() << endl;
      com << "* BYE Binc IMAP shutting down: " << e.getReason() << endl;
      session.setState(Session::LOGOUT);

    } catch (SyntaxException &e) {
      logger << "Caught SyntaxException: " << e.getReason() << endl;
      com << "* NO Syntax Error: " << e.getReason() << endl;
      recovery = true;

    } catch (InputException &e) {
      logger << "Input warning, shutting down: " << e.getReason() << endl;
      session.setState(Session::LOGOUT);

    } catch (BincException &e) {
      logger << "General exception: " << e.getReason() << endl;
      session.setState(Session::LOGOUT);

    } catch (...) {
      logger << "Anonymous exception caught" << endl;
      session.setState(Session::LOGOUT);
    }

    com.flushContent();

    // recover from syntax error. There will be trash in the input
    // buffer. We need to flush everything until we see an LF.
    if (recovery)
      try {
	while (com.readChar() != '\n');
      } catch (InputException &e) {
	logger << "IO warning: " << e.getReason() << endl;
	session.setState(Session::LOGOUT);
      }
  }

  logger << "Shutting down - read:" << session.getReadBytes() 
	 << " wrote:" << session.getWriteBytes() << endl;

  com.flushContent();
}
