/* -*- Mode: c++; -*- */
/*  --------------------------------------------------------------------
 *  Filename:
 *    src/operator-append.cc
 *  
 *  Description:
 *    Implementation of the APPEND command.
 *
 *  Authors:
 *    Andreas Aardal Hanssen <bincimap@andreas.hanssen.name>
 *
 *  Bugs:
 *
 *  ChangeLog:
 *
 *  --------------------------------------------------------------------
 *  Copyright 2002, 2003 Andreas Aardal Hanssen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *  --------------------------------------------------------------------
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <algorithm>
#include <string>

#include <fcntl.h>

#include "depot.h"
#include "io/io.h"
#include "mailbox/mailbox.h"
#include "operators.h"
#include "parsers/imap/recursivedescent/recursivedescent.h"
#include "pendingupdates.h"
#include "session.h"

using namespace ::std;
using namespace Binc;

//----------------------------------------------------------------------
AppendOperator::AppendOperator(void)
{
}

//----------------------------------------------------------------------
AppendOperator::~AppendOperator(void)
{
}

//----------------------------------------------------------------------
const string AppendOperator::getName(void) const
{
  return "APPEND";
}

//----------------------------------------------------------------------
int AppendOperator::getState(void) const
{
  return Session::AUTHENTICATED | Session::SELECTED;
}

//------------------------------------------------------------------------
void AppendOperator::process(Depot &depot,
			     Command &command) const
{
  IO &com = IOFactory::getInstance().get(1);
  IO &logger = IOFactory::getInstance().get(2);

  const string &srcmailbox = command.getMailbox();
  const string &canonmailbox = toCanonMailbox(srcmailbox);
  Mailbox *mailbox = NULL;

  if ((mailbox = depot.get(canonmailbox)) == NULL)
    throw OperatorException(Operator::NO, 
			    "[TRYCREATE] invalid destination mailbox "
			    + toImapString(srcmailbox));

  // Read number of characters in literal. Literal is required here.
  if (com.readChar() != '{')
    throw InputException("expected literal");
  
  string nr;
  while (1) {
    int c = com.readChar();
    if (c == -1)
      throw InputException("unexcepted EOF");
    if (c == '}')
      break;
    nr += (char) c;
  }

  int nchars = atoi(nr.c_str());
  if (nchars < 0)
    throw InputException("expected positive size of appended message");

  if (com.readChar() != '\r')
    throw InputException("expected CR");
  if (com.readChar() != '\n')
    throw InputException("expected LF");
  
  com << "+ go ahead with " << nchars << " characters" << endl;
  com.flushContent();
  
  char tempfile[64];
  strcpy(tempfile, "bincimap-append-XXXXXX");
  int fd = mkstemp(tempfile);
  if (fd == -1) {
    logger << "failed to create safe file: " 
	   << string(strerror(errno)) << endl;
    throw OperatorException(Operator::NO, "couldn't create message");
  }

  com.disableInputLimit();

  // Read all characters from client - but skip the CRs.
  for (int i = 0; i < nchars; ++i) {
    int c = com.readChar();
    if (c == -1) {
      close(fd);
      unlink(tempfile);
      throw InputException("unexpected EOF while appending");      
    } else {
      char cc = (char)c;
      if (cc != '\r')
	if (write(fd, &cc, 1) != 1) {
	  close(fd);
	  unlink(tempfile);
	  throw InputException("error writing while appending");
	}
    }
  }

  if (com.readChar() != '\r')
    throw InputException("expected CR");
  if (com.readChar() != '\n')
    throw InputException("expected CR");

  // sync file content
  // patched - EINVAL simply means that fsync isn't supported for the
  // given file and therefore we can ignore it. --cld 20030205
  if (fsync(fd) == -1 && errno != EINVAL) {
    close(fd);
    unlink(tempfile);
    throw OperatorException(Operator::NO, "error in fsync while appending");
  }
  close(fd);

  string flags;
  for (vector<string>::const_iterator i = command.flags.begin();
       i != command.flags.end(); ++i) {
    string t = *i;
    uppercase(t);
    if (t == "\\DRAFT") flags += "D";
    else if (t == "\\FLAGGED") flags += "F";
    else if (t == "\\ANSWERED") flags += "R";
    else if (t == "\\SEEN") flags += "S";
    else if (t == "\\DELETED") flags += "T";
  }

  sort(flags.begin(), flags.end());

  time_t passed_time = time(NULL);
  /*
    int mday, year, hour, minute, second;
    char month[4];

    sscanf(command.getDate().c_str(), "%2i-%3s-%4i %2i:%2i:%2i",
	   &mday, month, &year, &hour, &minute, &second);

    month[3] = '\0';
    string monthstr = month;
    lowercase(monthstr);

    struct tm mytm;
    mytm.tm_sec = second;
    mytm.tm_min = minute;
    mytm.tm_hour = hour;
    mytm.tm_year = year;
    mytm.tm_mday = mday;
    if (monthstr == "jan") mytm.tm_mon = 1;
    else if (monthstr == "feb") mytm.tm_mon = 2;
    else if (monthstr == "mar") mytm.tm_mon = 3;
    else if (monthstr == "apr") mytm.tm_mon = 4;
    else if (monthstr == "may") mytm.tm_mon = 5;
    else if (monthstr == "jun") mytm.tm_mon = 6;
    else if (monthstr == "jul") mytm.tm_mon = 7;
    else if (monthstr == "aug") mytm.tm_mon = 8;
    else if (monthstr == "sep") mytm.tm_mon = 9;
    else if (monthstr == "oct") mytm.tm_mon = 10;
    else if (monthstr == "nov") mytm.tm_mon = 11;
    else if (monthstr == "dec") mytm.tm_mon = 12;
    mytm.tm_isdst = -1;

    passed_time = mktime(&mytm);
  */

  char hostname[512];
  int hostnamelen = gethostname(hostname, sizeof(hostname));
  if (hostnamelen == -1 || hostnamelen == sizeof(hostname))
    strcpy(hostname, "localhost");

  string filename;

  // Make a number of attempts to create a unique file name in
  // new/. Restrictions in Maildir prevent us from storing a message
  // with flags, which breaks a SHOULD in section 6.3.11 of rfc2060.
  for (int attempt = 0; ; ++attempt) {
    BincStream ss;

    struct timeval tv;
    gettimeofday(&tv, NULL);

    ss << depot.mailboxToFilename(canonmailbox)
       << "/new/" << (unsigned int)passed_time
       << "." << getpid() << (attempt == 0 ? "" : ("_" + toString(attempt)))
       << "_" << tv.tv_usec << "." << hostname;

    filename = ss.str();

    // if link fails, we _can_ and _should_ try to create another file name
    if (link(tempfile, filename.c_str()) == -1) {
      if (errno != EEXIST) {
	logger << "APPEND failed: " << strerror(errno) << endl;
	unlink(tempfile);
	throw OperatorException(Operator::NO, 
				"[TRYCREATE] failed, couldn't link file");
      }
    } else break;

    ++attempt;
  }

  // open directory
  if ((fd = open(depot.mailboxToFilename(canonmailbox).c_str(),
		 0200000 /*O_DIRECTORY*/)) == -1)
    throw OperatorException(Operator::NO, "couldn't open \"" 
			    + depot.mailboxToFilename(canonmailbox) 
			    + "\": " + string(strerror(errno)));
  
  // sync directory (actually unnecessary on most unices, but it
  // doesn't hurt and I won't bother with detecting wether it's
  // necessary or not).
  // patched - EINVAL simply means that fsync isn't supported for the
  // given file and therefore we can ignore it. --cld 20030205
  if (fsync(fd) == -1 && errno != EINVAL) {
    close(fd);
    throw OperatorException(Operator::NO, 
			    "couldn't fsync cwd: " + string(strerror(errno)));
  }
  
  // close directory
  close(fd);

  // remove temp file
  unlink(tempfile);

  if (mailbox == depot.getSelected()) {
    sleep(1);
    pendingUpdates(mailbox,
		   PendingUpdates::EXISTS | PendingUpdates::RECENT, true);
  }
}

//----------------------------------------------------------------------
bool AppendOperator::parse(Command &c_in) const
{
  try {
    if (!expectThisString("APPEND"))
      return false;

    if (!expectSPACE())
      throw SyntaxException("Expected SPACE after APPEND");

    string mailbox;
    if (!expectMailbox(mailbox))
      throw SyntaxException("Expected mailbox after APPEND SPACE");
    c_in.setMailbox(mailbox);

    if (!expectSPACE())
      throw SyntaxException("Expected SPACE after APPEND SPACE mailbox");


    if (expectThisString("(")) {
      if (expectFlag(c_in.getFlags()))
	while (1) {
	  if (!expectSPACE())
	    break;
	  if (!expectFlag(c_in.getFlags()))
	    throw SyntaxException("expected a flag after the '('");
	}
      if (!expectThisString(")"))
	throw SyntaxException("expected a ')'");

      if (!expectSPACE())
	throw SyntaxException("expected a SPACE after the flag list");
    }

    string date;
    if (expectDateTime(date))
      if (!expectSPACE())
	throw SyntaxException("expected a SPACE after date_time");
    c_in.setDate(date);

  } catch (InputException &e) {
    e.addReason("in AppendOperator::parse");
    throw e;
  } catch (SyntaxException &e) {
    e.addReason("in AppendOperator::parse");
    throw e;
  }

  c_in.setName("APPEND");
  return true;
}
