/* -*- Mode: c++; -*- */
/*  --------------------------------------------------------------------
 *  Filename:
 *    src/operator-authenticate.cc
 *  
 *  Description:
 *    Implementation of the AUTHENTICATE command.
 *
 *  Authors:
 *    Andreas Aardal Hanssen <bincimap@andreas.hanssen.name>
 *
 *  Bugs:
 *
 *  ChangeLog:
 *
 *  --------------------------------------------------------------------
 *  Copyright 2002, 2003 Andreas Aardal Hanssen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *  --------------------------------------------------------------------
 */
#include <string>

#include "authenticate.h"
#include "depot.h"
#include "io/io.h"
#include "session.h"
#include "operators.h"
#include "parsers/imap/recursivedescent/recursivedescent.h"
#include "util/base64.h"
#include "util/convert.h"

using namespace ::std;
using namespace Binc;

//----------------------------------------------------------------------
AuthenticateOperator::AuthenticateOperator(void)
{
}

//----------------------------------------------------------------------
AuthenticateOperator::~AuthenticateOperator(void)
{
}

//----------------------------------------------------------------------
const string AuthenticateOperator::getName(void) const
{
  return "AUTHENTICATE";
}

//----------------------------------------------------------------------
int AuthenticateOperator::getState(void) const
{
  return Session::NONAUTHENTICATED;
}

//------------------------------------------------------------------------
void AuthenticateOperator::process(Depot &depot,
				   Command &command) const
{
  Session &session = Session::getInstance();
  IO &com = IOFactory::getInstance().get(1);

  string authtype = command.getAuthType();
  uppercase(authtype);

  string username;
  string password;

  session.globalconfig.setSection("Authentication");
  const bool allowplain
    = (session.globalconfig["allow plain auth in non ssl"] == "yes");

  // for now, we only support LOGIN.
  if (authtype == "LOGIN") {
    // we only allow this type of authentication over a plain
    // connection if it is passed as argument or given in the conf
    // file.
    if (!session.command.ssl && !allowplain)
      throw OperatorException(Operator::NO,
			      "disallowed when in plain-text."\
			      "Please connect using SSL/TLS");
    
    com << "+ " << base64encode("User Name") << endl;
    com.flushContent();

    char c;
    while ((c = com.readChar()) != '\n')
      username += c;
    
    if (username != "" && username[0] == '*')
      throw OperatorException(Operator::NO, "cancelled");
    
    com << "+ " << base64encode("Password") << endl;
    com.flushContent();
    
    while ((c = com.readChar()) != '\n')
      password += c;
    
    if (password != "" && password[0] == '*')
      throw OperatorException(Operator::NO, "cancelled");

    username = base64decode(username);
    password = base64decode(password);

  } else if (authtype == "PLAIN") {
    // we only allow this type of authentication over an SSL encrypted
    // connection.
    if (!session.command.ssl && !allowplain)
      throw OperatorException(Operator::NO,
			      "disallowed when in plain-text."\
			      "Please connect using SSL/TLS");

    com << "+" << endl;
    com.flushContent();
    
    string b64;
    char c;
    while ((c = com.readChar()) != '\n')
      if (c != '\r')
	b64 += c;

    if (b64.size() >= 1 && b64[0] == '*')
      throw OperatorException(Operator::NO, "cancelled");

    string plain = base64decode(b64);
    string::size_type pos;
    if ((pos = plain.find('\0')) == string::npos)
      throw OperatorException(Operator::BAD, "failed");
    
    plain = plain.substr(pos + 1);
    if ((pos = plain.find('\0')) == string::npos)
      throw OperatorException(Operator::BAD, "failed");

    username = plain.substr(0, pos);
    password = plain.substr(pos + 1);
  } else
    throw OperatorException(Operator::NO,
			    "unsupported authentication mechanism");

  putenv(strdup(("BINCIMAP_LOGIN=AUTHENTICATE+" + command.getTag()).c_str()));

  // the authenticate function calls a stub which does the actual
  // authentication. the function returns 0 (success), 1 (internal
  // error) or 2 (failed)
  switch (authenticate(depot,
		       (const string &)username,
		       (const string &)password)) {
  case 1:
    throw OperatorException(Operator::NO, "internal error");
  case 2:
    throw OperatorException(Operator::NO, "wrong user ID or password");
  }

  // auth was ok. go to logout state
  session.setState(Session::LOGOUT);
  throw OperatorException(Operator::NOTHING, "");
}


//----------------------------------------------------------------------
bool AuthenticateOperator::parse(Command &c_in) const
{
  try {
    if (!expectThisString("AUTHENTICATE"))
      return false;

    if (!expectSPACE())
      throw SyntaxException("Expected single SPACE after AUTHENTICATE");

    string authtype;
    if (!expectAtom(authtype))
      throw SyntaxException("Expected auth_type after AUTHENTICATE SPACE");

    if (!expectCRLF())
      throw SyntaxException("expected CRLF");

    c_in.setAuthType(authtype);

  } catch (InputException &e) {
    e.addReason("in AuthenticateOperator::parse");
    throw e;
  }

  c_in.setName("AUTHENTICATE");
  return true;
}
