/* -*- Mode: c++; -*- */
/*  --------------------------------------------------------------------
 *  Filename:
 *    src/operator-rename.cc
 *  
 *  Description:
 *    Implementation of the RENAME command.
 *
 *  Authors:
 *    Andreas Aardal Hanssen <bincimap@andreas.hanssen.name>
 *
 *  Bugs:
 *
 *  ChangeLog:
 *
 *  --------------------------------------------------------------------
 *  Copyright 2002, 2003 Andreas Aardal Hanssen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *  --------------------------------------------------------------------
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string>
#include <iostream>

#include <dirent.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "mailbox/mailbox.h"
#include "io/io.h"
#include "parsers/storage/storage.h"
#include "util/convert.h"

#include "parsers/imap/recursivedescent/recursivedescent.h"

#include "session.h"
#include "depot.h"
#include "operators.h"

using namespace ::std;
using namespace Binc;

//----------------------------------------------------------------------
RenameOperator::RenameOperator(void)
{
}

//----------------------------------------------------------------------
RenameOperator::~RenameOperator(void)
{
}

//----------------------------------------------------------------------
const string RenameOperator::getName(void) const
{
  return "RENAME";
}

//----------------------------------------------------------------------
int RenameOperator::getState(void) const
{
  return Session::AUTHENTICATED | Session::SELECTED;
}

//------------------------------------------------------------------------
void RenameOperator::process(Depot &depot,
			     Command &command) const
{
  const string &srcmailbox = command.getMailbox();

  const string &canonmailbox = toCanonMailbox(srcmailbox);
  const string &canondestmailbox = toCanonMailbox(command.getNewMailbox());

  // renaming INBOX should actually create the destination mailbox,
  // move over all the messages and then leave INBOX empty.
  if (canonmailbox == "INBOX")
    throw OperatorException(Operator::NO, "unsupported");

  if (canondestmailbox == "INBOX")
    throw OperatorException(Operator::NO, "disallowed by rule");

  try {
    depot.renameMailbox(canonmailbox, canondestmailbox);
  } catch (StorageException &e) {
    throw OperatorException(Operator::NO, e.getReason());
  }
}

//----------------------------------------------------------------------
bool RenameOperator::parse(Command &c_in) const
{
  string mailbox;
  string newmailbox;
  
  try {
    if (!expectThisString("RENAME"))
      return false;

    if (!expectSPACE())
      throw SyntaxException("expected SPACE");

    if (!expectMailbox(mailbox))
      throw SyntaxException("expected mailbox");    

    if (!expectSPACE())
      throw SyntaxException("expected SPACE");

    if (!expectMailbox(newmailbox))
      throw SyntaxException("expected mailbox");
    if (!expectCRLF())
      throw SyntaxException("expected CRLF");

  } catch (InputException &e) {
    e.addReason("in RenameOperator::parse");
    throw e;
  }

  c_in.setName("RENAME");
  c_in.setMailbox(mailbox);
  c_in.setNewMailbox(newmailbox);

  return true;
}
