/* -*- Mode: c++; -*- */
/*  --------------------------------------------------------------------
 *  Filename:
 *    src/operator-select.cc
 *  
 *  Description:
 *    Implementation of the SELECT command.
 *
 *  Authors:
 *    Andreas Aardal Hanssen <bincimap@andreas.hanssen.name>
 *
 *  Bugs:
 *
 *  ChangeLog:
 *
 *  --------------------------------------------------------------------
 *  Copyright 2002, 2003 Andreas Aardal Hanssen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *  --------------------------------------------------------------------
 */
#include <string>

#include "depot.h"
#include "io/io.h"
#include "mailbox/mailbox.h"
#include "mailbox/maildir/maildir.h"
#include "operators.h"
#include "parsers/imap/recursivedescent/recursivedescent.h"
#include "parsers/storage/storage.h"
#include "pendingupdates.h"
#include "session.h"
#include "util/convert.h"

using namespace ::std;
using namespace Binc;

//----------------------------------------------------------------------
SelectOperator::SelectOperator(void)
{
}

//----------------------------------------------------------------------
SelectOperator::~SelectOperator(void)
{
}

//----------------------------------------------------------------------
const string SelectOperator::getName(void) const
{
  return "SELECT";
}

//----------------------------------------------------------------------
int SelectOperator::getState(void) const
{
  return Session::NONAUTHENTICATED
    | Session::AUTHENTICATED
    | Session::SELECTED;
}

//------------------------------------------------------------------------
void SelectOperator::process(Depot &depot,
			     Command &command) const
{
  Session &session = Session::getInstance();
  IO &com = IOFactory::getInstance().get(1);
  IO &logger = IOFactory::getInstance().get(2);

  const bool examine = (command.getName() == "EXAMINE");

  const string &srcmailbox = command.getMailbox();
  const string &canonmailbox = toCanonMailbox(srcmailbox);

  Mailbox *mailbox = depot.getSelected();
  if (mailbox != NULL)
    mailbox->imapClose();

  mailbox = depot.get(canonmailbox);
  if (mailbox == NULL)
    throw OperatorException(Operator::NO,
			    "invalid mailbox: " + toImapString(canonmailbox));

  try {
    mailbox->imapSelect(depot.mailboxToFilename(canonmailbox));
  } catch (BincException &e) {
    throw OperatorException(Operator::NO,
			    "unable to open " + toImapString(srcmailbox) 
			    + ", " + e.getReason());
  }

  // find first unseen
  int unseen = -1;
  Mailbox::iterator e = mailbox->end();
  for (Mailbox::iterator i = mailbox->begin(); i != e; ++i) {
    Message & message = *i;

    if (unseen == -1 && ((message.getStdFlags() & Message::F_SEEN) == 0)) {
      unseen = i.getSqnr();
      break;
    }
  }

  // show pending updates with only exists and recent response. do not re-scan.
  pendingUpdates(mailbox, PendingUpdates::EXISTS | PendingUpdates::RECENT,
		 false);

  // unseen
  if (unseen != -1)
    com << "*" << " OK [UNSEEN " << unseen << "] Message "
	<< unseen << " is first unseen" << endl;

  // uidvalidity
  com << "*" << " OK [UIDVALIDITY " << mailbox->getUidValidity() << "]"
      << endl;

  // uidnext
  com << "*" << " OK [UIDNEXT " << toString(mailbox->getUidNext()) << "] "
      << toString(mailbox->getUidNext()) << " is the next UID" << endl;

  // flags
  com << "*" 
      << " FLAGS (\\Answered \\Flagged \\Deleted \\Recent \\Seen \\Draft)"
      << endl;

  // permanentflags
  com << "*" 
      << " OK [PERMANENTFLAGS (\\Answered \\Flagged \\Deleted "
      << "\\Seen \\Draft)] Limited" 
      << endl;

  session.setState(Session::SELECTED);
  depot.setSelected(mailbox);

  if (examine) mailbox->setReadOnly();

  logger.setLogPrefix(session.getUserID() + "@" + session.getIP()
		      + ":" + srcmailbox);

  throw OperatorException(Operator::OK, examine ?
			  "[READ-ONLY]" : "[READ-WRITE]");
}

//----------------------------------------------------------------------
bool SelectOperator::parse(Command &c_in) const
{
  string mailbox;
  bool examine;
  
  try {
    if (expectThisString("EXAMINE"))
      examine = true;
    else if (expectThisString("SELECT"))
      examine = false;
    else
      return false;

    if (!expectSPACE())
      throw SyntaxException("expected SPACE");

    if (!expectMailbox(mailbox))
      throw SyntaxException("expected mailbox");    
    if (!expectCRLF())
      throw SyntaxException("expected CRLF");

  } catch (InputException &e) {
    e.addReason("in SelectOperator::parse");
    throw e;
  }

  if (examine)
    c_in.setName("EXAMINE");
  else
    c_in.setName("SELECT");
  c_in.setMailbox(mailbox);

  return true;
}
