/* -*- Mode: c++; -*- */
/*  --------------------------------------------------------------------
 *  Filename:
 *    src/operator-store.cc
 *  
 *  Description:
 *    Implementation of the STORE command.
 *
 *  Authors:
 *    Andreas Aardal Hanssen <bincimap@andreas.hanssen.name>
 *
 *  Bugs:
 *
 *  ChangeLog:
 *
 *  --------------------------------------------------------------------
 *  Copyright 2002, 2003 Andreas Aardal Hanssen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *  --------------------------------------------------------------------
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string>
#include <iostream>
    
#include "parsers/imap/imapparser.h"
#include "mailbox/mailbox.h"
#include "pendingupdates.h"
#include "io/io.h"

#include "parsers/imap/recursivedescent/recursivedescent.h"

#include "session.h"
#include "depot.h"
#include "operators.h"

using namespace ::std;
using namespace Binc;

//----------------------------------------------------------------------
StoreOperator::StoreOperator(void)
{
}

//----------------------------------------------------------------------
StoreOperator::~StoreOperator(void)
{
}

//----------------------------------------------------------------------
const string StoreOperator::getName(void) const
{
  return "STORE";
}

//----------------------------------------------------------------------
int StoreOperator::getState(void) const
{
  return Session::SELECTED;
}

//------------------------------------------------------------------------
void StoreOperator::process(Depot &depot, Command &command) const
{
  Mailbox *mailbox = depot.getSelected();
  if (mailbox == NULL)
    throw CriticalException("store when no mailbox is selected");

  // mask all passed flags together
  unsigned int newflags = (unsigned int) Message::F_NONE;
  vector<string>::const_iterator f_i = command.flags.begin();
  while (f_i != command.flags.end()) {
    if (*f_i == "\\Deleted") newflags |= Message::F_DELETED;
    if (*f_i == "\\Answered") newflags |= Message::F_ANSWERED;
    if (*f_i == "\\Seen") newflags |= Message::F_SEEN;
    if (*f_i == "\\Draft") newflags |= Message::F_DRAFT;
    if (*f_i == "\\Flagged") newflags |= Message::F_FLAGGED;
    ++f_i;
  }

  // pass through all messages
  unsigned int mode;
  if (command.getUidMode())
    mode = Mailbox::UID_MODE;
  else
    mode = Mailbox::SQNR_MODE;
  Mailbox::iterator i = mailbox->begin(command.bset,
				       Mailbox::SKIP_EXPUNGED | mode);
  Mailbox::iterator e = mailbox->end();
  for (;i != e; ++i) {
    Message & message = *i;

    // get and reset the old flags
    unsigned int oldflags = (unsigned int) message.getStdFlags();
    unsigned int flags = oldflags;
    
    bool recent = (flags & Message::F_RECENT) != 0;
    flags &= (~Message::F_RECENT);

    // add, remove or set flags
    if (command.getMode()[0] == '+') flags |= newflags;
    else if (command.getMode()[0] == '-') flags &= ~newflags;
    else flags = newflags;

    // set new flags, even if they weren't changed.
    if (recent) flags |= Message::F_RECENT;
    message.resetStdFlags();
    message.setStdFlag(flags);
  }

  // commit flag changes to mailbox (might change mailbox)
  mailbox->updateFlags();

  // check mailbox for updates, and report them
  if (command.getMode().find(".SILENT") != string::npos)
    pendingUpdates(mailbox,
		   PendingUpdates::EXISTS
		   | PendingUpdates::RECENT, true);
  else
    pendingUpdates(mailbox,
		   PendingUpdates::EXISTS 
		   | PendingUpdates::RECENT 
		   | PendingUpdates::FLAGS, true);
}

//----------------------------------------------------------------------
bool StoreOperator::parse(Command & c_in) const
{
  IO &com = IOFactory::getInstance().get(1);

  try {
    bool uidmode = false;
    if (expectThisString("UID")) {
      uidmode = true;
      if (!expectSPACE())
	throw SyntaxException("expected SPACE");
    }

    if (!expectThisString("STORE")) {
      if (uidmode)
	com.unReadChar("UID ");
      return false;
    }

    if (!expectSPACE())
      throw SyntaxException("expected SPACE");

    if (!expectSet(c_in.getSet()))
      throw SyntaxException("expected Set");

    if (!expectSPACE())
      throw SyntaxException("expected SPACE");

    string mode;
    if (expectThisString("+"))
      mode = "+";
    else if (expectThisString("-"))
      mode = "-";

    if (!expectThisString("FLAGS"))    
      throw SyntaxException("expected FLAGS");
    else
      mode += "FLAGS";

    if (expectThisString(".SILENT"))
      mode += ".SILENT";

    c_in.setMode(mode);

    if (!expectSPACE())
      throw SyntaxException("expected SPACE");

    bool paren = false;
    if (expectThisString("("))
      paren = true;
     
    if (expectFlag(c_in.getFlags()))
      while (1) {
	if (!expectSPACE())
	  break;

	if (!expectFlag(c_in.getFlags()))
	  throw SyntaxException("expected flag after SPACE");
      }

    if (paren)
      if (!expectThisString(")"))
	throw SyntaxException("expected )");
    if (!expectCRLF())
      throw SyntaxException("expected CRLF");
    if (uidmode) c_in.setUidMode();
    c_in.setName("STORE");
    return true;

  } catch (InputException &e) {
    e.addReason("in StoreOperator::parse");
    throw e;
  }
  return false;
}
