/* -*- Mode: c++; -*- */
/*  --------------------------------------------------------------------
 *  Filename:
 *    src/parsers/storage/storage.cc
 *  
 *  Description:
 *    Implementation of the Storage class.
 *
 *  Authors:
 *    Andreas Aardal Hanssen <bincimap@andreas.hanssen.name>
 *
 *  Bugs:
 *
 *  ChangeLog:
 *
 *  --------------------------------------------------------------------
 *  Copyright 2002, 2003 Andreas Aardal Hanssen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *  --------------------------------------------------------------------
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "storage.h"
#include "../../exceptions.h"
#include "../../io/io.h"
#include "../../util/convert.h"
#include "../../util/base64.h"
#include <string>
#include <iostream>
#include <iomanip>
#include <map>

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

using namespace ::std;
using namespace Binc;

//------------------------------------------------------------------------
Storage *Binc::internal__storage;

//------------------------------------------------------------------------
Binc::Storage::Storage(void) : internalStorage()
{
  internalStorageMap = NULL;
  savemode = STORAGE_MINIMIZE;
}

//------------------------------------------------------------------------
Binc::Storage::~Storage(void)
{
}

//------------------------------------------------------------------------
void Binc::Storage::load(const string &filename)
{
  FILE *fp = fopen(filename.c_str(), "r");
  if (fp == NULL) {
    string reason = "can't load " + filename + ", ";
    reason += strerror(errno);
    throw StorageException(reason);
  }

  storageparserin = fp;
  
  storage_charnr = 0;
  storage_linenr = 1;

  internal__storage = this;

  storageparserparse();
  aliastotext.clear();
  texttoalias.clear();

  fclose(fp);
  storageparserin = NULL;
  internal__storage = NULL;
}

namespace {
  //----------------------------------------------------------------------
  inline const string &createsuggestion(const string &s)
  {
    static const string abbrev = 
      "abcdefghijklmnopqrstuvwxyz"\
      "ABCDEFGHIJKLMNOPQRSTUVWXYZ"\
      "0123456789";

    static string olds = "";
    static int counter = -1;
    
    if (s != olds) {
      olds = s;
      counter = -1;
    }
    
    // only try to create one hundred suggestions
    if (counter == sizeof(abbrev) - 1)
      return s;
    
    static string buf;
    buf = "";
    if (counter == -1) {
      counter = 0;
      buf = s[0];
    } else {
      buf = abbrev[counter];
    }
     
    return buf;
  }

  //----------------------------------------------------------------------
  inline bool islabelchar(char c)
  {
    switch (c) {
    case 0x2D: case 0x2E:
    case 0x30: case 0x31: case 0x32: case 0x33: case 0x34:
    case 0x35: case 0x36: case 0x37: case 0x38: case 0x39: 
    case 0x40: case 0x41: case 0x42: case 0x43: case 0x44:
    case 0x45: case 0x46: case 0x47: case 0x48: case 0x49:
    case 0x4A: case 0x4B: case 0x4C: case 0x4D: case 0x4E:
    case 0x4F: case 0x50: case 0x51: case 0x52: case 0x53:
    case 0x54: case 0x55: case 0x56: case 0x57: case 0x58:
    case 0x59: case 0x5A:
      //    case 0x5B:
      //    case 0x5C:
      //    case 0x5D:
    case 0x5E: case 0x5F:
      //    case 0x60:
    case 0x61: case 0x62: case 0x63: case 0x64: case 0x65:
    case 0x66: case 0x67: case 0x68: case 0x69: case 0x6A:
    case 0x6B: case 0x6C: case 0x6D: case 0x6E: case 0x6F:
    case 0x70: case 0x71: case 0x72: case 0x73: case 0x74:
    case 0x75: case 0x76: case 0x77: case 0x78: case 0x79:
    case 0x7A:
      return true;
    default: return false;
    }
  }

  //----------------------------------------------------------------------
  inline const string &encode(const string &s)
  {
    static string output;

    for (string::const_iterator i = s.begin(); i != s.end(); ++i)
      if ((unsigned char)*i < 32 || (unsigned char)*i > 127) {
	output += "<";
	output += toHex(s);
	output += ">";
	return output;
      }

    for (string::const_iterator i = s.begin(); i != s.end(); ++i)
      if (!islabelchar(*i)) {
	output = "\"";
	char c;
	for (string::const_iterator i = s.begin(); i != s.end(); ++i)
	  switch ((c = *i)) {
	  case '\"': output += "\\\""; break;
	  case '\\': output += "\\\\"; break;
	  default: output += c; break;
	  }
	output += "\"";
	return output;
      }

    return output = s;
  }

  //----------------------------------------------------------------------
  inline void writeData(int fd, const string &out)
  {
    using namespace Binc;
    static string buffer;

    buffer += out;
    if (buffer.size() >= 8192 || out == "") {
      int written = 0;
      while (1) {
	int retval = write(fd, buffer.c_str() + written, buffer.length() - written);
	if (retval == (int) buffer.length())
	  break;
	
	if (retval == -1) {
	  if (errno == EINTR) continue;
	  throw StorageException("error writing: " + string(strerror(errno)));
	}
	
	written += retval;
      }

      buffer = "";
    }
  }
}

//------------------------------------------------------------------------
void Binc::Storage::save(const string &filename)
{
  map<string, string> aliastotext;
  map<string, string> texttoalias;

  // if we are asked to use MINIMIZE, then we first detect all keys
  // and section names in the whole object. we then find the smallest
  // prefices available that can replace each of these keys and names
  // (called labels), and store them in two maps. this is of course
  // only done if a label is used more than once.
  if (savemode == STORAGE_MINIMIZE) {
    map<string, int> labels;

    // find all labels
    for (map<string, multimap<string, string> >::iterator i 
	   = internalStorage.begin(); i != internalStorage.end(); ++i) {
      labels[i->first]++;
      multimap<string, string> &m = i->second;
      for (multimap<string, string>::iterator j = m.begin(); j != m.end(); ++j)
	labels[j->first]++;
    }
        
    // create a map from label names to aliases
    for (map<string, int>::iterator i = labels.begin(); 
	 i != labels.end(); ++i) {
      if (i->second == 1)
	continue;
      
      string s;
      bool match = false;
      while ((s = createsuggestion(i->first)) != "")
	if (aliastotext.find(s) == aliastotext.end()) {
	  texttoalias[i->first] = s;
	  aliastotext[s] = i->first;
	  match = true;
	  break;
	}
    }
  }
   
  // create a safe file name
  string tpl = filename + "XXXXXX";
  char *ftemplate = new char[tpl.length() + 1];
  if (ftemplate == NULL)
    throw StorageException("out of memory, couldn't create safe filename");
  strcpy(ftemplate, tpl.c_str());
  int fd = mkstemp(ftemplate);

  // write aliases if we're asked to and there are any
  if (savemode == STORAGE_MINIMIZE)
    for (map<string, string>::iterator i = aliastotext.begin(); 
	 i != aliastotext.end(); ++i)      
      if (i->first != i->second.c_str()) {
	string t = "?";
	t += i->first;
	t += ":";
	t += i->second;
	t += ";";
	writeData(fd, t);
      }

  // don't need this more
  aliastotext.clear();

  // write sections
  for (map<string, multimap<string, string> >::iterator i 
	 = internalStorage.begin(); i != internalStorage.end(); ++i) {

    string s = i->first;

    // find alias
    if (savemode == STORAGE_MINIMIZE)
      if (texttoalias.find(s) != texttoalias.end())
	s = texttoalias[s];

    // write the section title
    if (savemode == STORAGE_MINIMIZE)
      writeData(fd, s + "{");
    else {
      if (i != internalStorage.begin())
	writeData(fd, "\n");
      writeData(fd, s + " {\n");
    }

    // write all members of the section in an appropriate
    // representation.
    multimap<string, string> &m = i->second;
    for (multimap<string, string>::iterator j = m.begin(); 
	 j != m.end(); ++j) {

      // find, convert to alias, and perhaps also encode the key.
      string s = j->first;
      if (savemode == STORAGE_MINIMIZE && texttoalias.find(s) != texttoalias.end())
	s = texttoalias[s];

      // find and perhaps also encode the value.
      string t = j->second;

      // write the key=value pair
      if (savemode == STORAGE_MINIMIZE) {
	string tmp2;
	if (j != m.begin()) tmp2 += ",";
	tmp2 += encode(s);
	tmp2 += "=";
	tmp2 += encode(t);
	writeData(fd, tmp2);
      } else {
	string tmp2;
	tmp2 += (j != m.begin()) ? ",\n\t" : "\t";
	tmp2 += encode(s);
	tmp2 += "=";
	tmp2 += encode(t);
	writeData(fd, tmp2);
      }
    }

    // write the end of the section
    writeData(fd, savemode == STORAGE_MINIMIZE ? "}" : "\n}\n");
  }

  // this is the equivalent of "flush"
  writeData(fd, "");

  // sync file contents
  // patched - EINVAL simply means that fsync isn't supported for the
  // given file and therefore we can ignore it. --cld 20030205
  if (fsync(fd) == -1 && errno != EINVAL) {
    close(fd);
    throw StorageException("couldn't fsync " + string(ftemplate)
			   + ": "  + string(strerror(errno)));
  }

  // close file
  if (close(fd) != 0)
    throw StorageException("couldn't fsync " + string(ftemplate)
			   + ": " +  string(strerror(errno)));

  // open directory
  string dir;
  if (filename.rfind('/') != string::npos) {
    dir = filename.substr(0, filename.rfind('/'));
  } else
    dir = ".";
  
  if ((fd = open(dir.c_str(), 0200000 /*O_DIRECTORY*/)) == -1)
    throw StorageException("couldn't open cwd: " + string(strerror(errno)));

  // sync directory (actually unnecessary on most unices, but it
  // doesn't hurt and I won't bother with detecting wether it's
  // necessary or not)
  if (fsync(fd) == -1 && errno != EINVAL) {
    close(fd);
    throw StorageException("couldn't fsync cwd: " + string(strerror(errno)));
  }

  // close directory
  if (close(fd) != 0)
    throw StorageException("couldn't fsync " + string(ftemplate)
			   + ": " +  string(strerror(errno)));

  // rename the file
  if (rename(ftemplate, filename.c_str()) != 0)
    throw StorageException("couldn't rename " + string(ftemplate)
			   + ": " + string(strerror(errno)));

  // is it necessary to sync the file/dir after the rename? well, if
  // the machine crashes at this point, the user either has an old
  // cache file or a new one, and nobody cares which is there,
  // really. :)

  delete ftemplate;
}

//------------------------------------------------------------------------
void Binc::Storage::setSaveMode(int i) throw (StorageException &)
{
  switch (i) {
  case STORAGE_MINIMIZE: 
  case STORAGE_PRETTIFY: savemode = i; break;
  default:
    throw StorageException("invalid save mode");
  }
}
