/* -*- Mode: c++; -*- */
/*  --------------------------------------------------------------------
 *  Filename:
 *    src/parsers/storage/storage.h
 *  
 *  Description:
 *    Declaration of the Storage class.
 *
 *  Authors:
 *    Andreas Aardal Hanssen <bincimap@andreas.hanssen.name>
 *
 *  Bugs:
 *
 *  ChangeLog:
 *
 *  --------------------------------------------------------------------
 *  Copyright 2002, 2003 Andreas Aardal Hanssen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *  --------------------------------------------------------------------
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifndef storage_headers_h_included
#define storage_headers_h_included
#include <string>
#include <map>

#include "../../exceptions.h"

#include <stdio.h>

namespace Binc {

  //------------------------------------------------------------------------
  class Storage {

  protected:
    mutable std::map<std::string, std::multimap<std::string, std::string> > 
    internalStorage;

    mutable std::multimap<std::string, std::string> *internalStorageMap;

    mutable std::multimap<std::string,
			  std::string>::const_iterator 
    internalEntryIterator;
    mutable std::map<std::string, 
		     std::multimap<std::string, std::string> >::const_iterator
    internalSectionIterator;

    mutable std::string cursection;

    int savemode;

  public:
    //-- access methods, get
    
    virtual const std::string &getSection(void) { return cursection; }

    virtual bool getFirstSection(std::string &section) const
    {
      internalSectionIterator = internalStorage.begin();
      if (internalSectionIterator == internalStorage.end())
	return false;
      section = internalSectionIterator->first;
      return true;    
    }

    virtual bool getNextSection(std::string &section) const
    {
      if (internalSectionIterator == internalStorage.end())
	return false;
      ++internalSectionIterator;
      if (internalSectionIterator == internalStorage.end())
	return false;
      section = internalSectionIterator->first;
      return true;    
    }

    virtual bool getFirstEntry(std::string &key, std::string &value) const
    {
      internalEntryIterator = internalStorage[cursection].begin();
      if (internalEntryIterator == internalStorage[cursection].end())
	return false;
      key = internalEntryIterator->first;
      value = internalEntryIterator->second;
      return true;
    }

    virtual bool getNextEntry(std::string &key, std::string &value) const
    {
      if (internalEntryIterator == internalStorage[cursection].end())
	return false;
      ++internalEntryIterator;
      if (internalEntryIterator == internalStorage[cursection].end())
	return false;
      key = internalEntryIterator->first;
      value = internalEntryIterator->second;
      return true;
    }

    virtual bool hasKey(const std::string &k_in) const {
      return (internalStorage[cursection].find(k_in)
	      != internalStorage[cursection].end());
    }

    virtual const std::string &operator [] (const std::string &k_in) const
    {
      const static std::string NIL = "";
      const std::multimap<std::string, std::string> &m = *internalStorageMap;

      if (m.find(k_in) == m.end())
	return NIL;

      return m.lower_bound(k_in)->second;
    }

    virtual std::string &get(const std::string &k_in)
    {
      std::multimap<std::string, std::string> &m = *internalStorageMap;
      
      if (m.find(k_in) == m.end())
	m.insert(make_pair(k_in, std::string("")));
      
      return m.lower_bound(k_in)->second;
    }

    //-- access methods, set

    inline virtual void setSection(const std::string &s_in)
    {
      cursection = s_in;
      internalStorageMap = &internalStorage[s_in];
    }

    inline virtual void insert(const std::string &k_in,
			       const std::string &v_in)
    {
      if (internalStorageMap == NULL)
	throw StorageException("attempted to insert {" + k_in + "=" + v_in + "} without selecting section");
      
      std::multimap<std::string, std::string>::iterator pos;
      while ((pos = internalStorageMap->find(k_in)) != internalStorageMap->end())
	internalStorageMap->erase(pos);

      internalStorageMap->insert(make_pair(k_in, v_in));
    }

    //-- construction and destruction

    enum {STORAGE_MINIMIZE, STORAGE_PRETTIFY};

    virtual void setSaveMode(int i) throw (StorageException &);

    virtual void load(const std::string &filename);
    virtual void save(const std::string &filename);

    Storage();
    virtual ~Storage();
  };

  // Semi global declarations that have to go here because of lex &
  // yacc. If anyone has a better way to do this, I'm all ears.
  int storageparserparse();

  extern FILE *storageparserin;
  extern Storage *internal__storage;

  extern std::map<std::string, std::string> aliastotext;
  extern std::map<std::string, std::string> texttoalias;

  extern int storage_charnr;
  extern int storage_linenr;
}

#endif
